
import unittest
import os
import sys
from itertools import zip_longest
import typing as tp
import itertools as it
import string
import cmath
import sqlite3
import contextlib
import tempfile
import time
import datetime
from pathlib import Path

import numpy as np
import pytest


from static_frame import TypeBlocks
from static_frame.core.container import ContainerOperand
from static_frame.core.frame import Frame
from static_frame.core.index_base import IndexBase
from static_frame.core.util import PathSpecifier


# for running with coverage
# pytest -s --color no --disable-pytest-warnings --cov=static_frame --cov-report html static_frame/test
# for running with native traveback
# pytest -s --color no --disable-pytest-warnings --tb=native


skip_win = pytest.mark.skipif(
        sys.platform.startswith('win'),
        reason='Windows default dtypes'
        )

skip_pylt37 = pytest.mark.skipif(
        sys.version_info < (3, 7),
        reason='Python earlier than 3.7'
        )

skip_linux_no_display = pytest.mark.skipif(
        sys.platform == 'linux' and 'DISPLAY' not in os.environ,
        reason='No display available'
        )

#-------------------------------------------------------------------------------
class Timer():

    def __init__(self, label: str = ''):
        self._start = time.time()
        self._label = label

    def __call__(self) -> float:
        return time.time() - self._start

    def __str__(self) -> str:
        if self._label:
            return f'{self._label}: {datetime.timedelta(seconds=self.__call__())}'
        return str(datetime.timedelta(seconds=self.__call__()))

#-------------------------------------------------------------------------------

@contextlib.contextmanager
def temp_file(suffix: tp.Optional[str] = None,
        path: bool = False
        ) -> tp.Iterator[PathSpecifier]:
    try:
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as f:
            tmp_name = f.name
        if path:
            yield Path(tmp_name)
        else:
            yield tmp_name
    finally:
        if os.path.exists(tmp_name):
            try:
                os.unlink(tmp_name)
            except PermissionError: # happens on Windows sometimes
                pass

class TestCase(unittest.TestCase):
    '''
    TestCase specialized for usage with StaticFrame
    '''

    @staticmethod
    def get_arrays_a() -> tp.Iterator[np.ndarray]:
        '''
        Return sample array suitable for TypeBlock creation, testing. Unique values required.
        '''

        a1 = np.array([[1, 2, 3], [10, 20, 30], [100, 200, 300]])
        a1.flags.writeable = False

        a2 = np.array([[4], [5], [6]])
        a2.flags.writeable = False

        a3 = np.array([[None, 'a', None], ['q', 'x', 'c'], ['f', 'y', 'e']])
        a3.flags.writeable = False

        a4 = np.array([1.2, np.nan, 30.5])
        a4.flags.writeable = False

        for arrays in it.permutations((a1, a2, a3, a4)):
            yield arrays


    @staticmethod
    def get_arrays_b() -> tp.Iterator[np.ndarray]:
        '''
        Return sample array suitable for TypeBlock creation, testing. Many NaNs.
        '''

        a1 = np.array([[1, 2, 3], [10, 20, 30], [100, 200, 300]])
        a1.flags.writeable = False

        a2 = np.array([[4], [5], [6]])
        a2.flags.writeable = False

        a3 = np.array([[None, 'a', None], [None, None, 'c'], ['f', None, 'e']])
        a3.flags.writeable = False

        a4 = np.array([np.nan, np.nan, np.nan])
        a4.flags.writeable = False

        for arrays in it.permutations((a1, a2, a3, a4)):
            yield arrays


    @staticmethod
    def get_letters(*slice_args: tp.Optional[int]) -> tp.Iterator[str]:
        for letter in string.ascii_lowercase[slice(*slice_args)]:
            yield letter

    @staticmethod
    def get_test_input(file_name: str) -> str:
        # input dir should be a sibling of this module
        fp_module = os.path.join(os.getcwd(), __file__)
        fp = os.path.join(os.path.dirname(fp_module), 'input', file_name)
        if not os.path.isfile(fp):
            raise RuntimeError('file not found', fp)
        return fp


    @staticmethod
    def get_containers() -> tp.Iterator[tp.Type[ContainerOperand]]:

        def yield_sub(cls: tp.Type[ContainerOperand]) -> tp.Iterator[tp.Type[ContainerOperand]]:
            for cls in cls.__subclasses__():
                if cls is not IndexBase:
                    yield cls
                if issubclass(cls, ContainerOperand):
                    yield from yield_sub(cls)

        yield from yield_sub(ContainerOperand)

    @staticmethod
    def get_test_db_a() -> sqlite3.Connection:
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''CREATE TABLE events
             (date text, identifier text, value real, count int)''')
        for identifier in ('a1', 'b2'):
            for date in ('2006-01-01', '2006-01-02'):
                c.execute(f"INSERT INTO events VALUES ('{date}','{identifier}',12.5,8)")
        conn.commit()
        return conn

    @staticmethod
    def get_test_db_b() -> sqlite3.Connection:
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''CREATE TABLE events
             (idx int, date text, identifier text, value real, count int)''')

        count = 0
        for identifier in ('a1', 'b2'):
            for date in ('2006-01-01', '2006-01-02'):
                c.execute(f"INSERT INTO events VALUES ({count}, '{date}','{identifier}',12.5,8)")
                count += 1
        conn.commit()
        return conn

    @staticmethod
    def get_test_db_c() -> sqlite3.Connection:
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''CREATE TABLE events
             ("'date' 'from'" text, "'date' 'to'" text, "'value' 'a'" real, "'value' 'b'" int)''')

        for identifier in ('a1', 'b2'):
            for date in ('2006-01-01', '2006-01-02'):
                c.execute(f"INSERT INTO events VALUES ('{date}','{identifier}',12.5,8)")
        conn.commit()
        return conn

    @staticmethod
    def get_test_db_d() -> sqlite3.Connection:
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''CREATE TABLE events
             ("'id' 'id'" text, "'date' 'from'" text, "'date' 'to'" text, "'value' 'a'" real, "'value' 'b'" int)''')

        count = 0
        for identifier in ('a1', 'b2'):
            for date in ('2006-01-01', '2006-01-02'):
                c.execute(f"INSERT INTO events VALUES ({count}, '{date}','{identifier}',12.5,8)")
                count += 1
        conn.commit()
        return conn

    @staticmethod
    def get_test_db_e() -> sqlite3.Connection:
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''CREATE TABLE events
             (date text, identifier text, value real, count int)''')
        count = 20
        for date in ('2006-01-01', '2006-01-02'):
            for identifier in ('a1', 'b2'):
                c.execute(f"INSERT INTO events VALUES ('{date}','{identifier}',12.5,{count})")
                count += 1
        conn.commit()
        return conn

    @staticmethod
    def get_test_db_f() -> sqlite3.Connection:
        conn = sqlite3.connect(':memory:')
        c = conn.cursor()
        c.execute('''CREATE TABLE events
             (count int, date text, identifier text, value real)''')
        count = 20
        for date in ('2006-01-01', '2006-01-02'):
            for identifier in ('a1', 'b2'):
                c.execute(f"INSERT INTO events VALUES ({count},'{date}','{identifier}',12.5)")
                count += 1
        conn.commit()
        return conn

    #---------------------------------------------------------------------------

    def assertEqualWithNaN(self,
            v1: object,
            v2: object,
            ) -> None:
        if ((isinstance(v1, float) or isinstance(v1, np.floating))
                and np.isnan(v1)
                and (isinstance(v2, float) or isinstance(v2, np.floating))
                and np.isnan(v2)
                ):
            return

        if ((isinstance(v1, complex) or isinstance(v1, np.complexfloating))
                and cmath.isnan(v1)
                and (isinstance(v2, complex) or isinstance(v1, np.complexfloating))
                and cmath.isnan(v2)  # type: ignore
                ):
            return
        return self.assertEqual(v1, v2)


    def assertAlmostEqualArray(self, a1: np.ndarray, a2: np.ndarray) -> None:
        # NaNs are treated as equal
        np.testing.assert_allclose(a1, a2)
        # np.testing.assert_array_almost_equal(a1, a2, decimal=5)

    def assertTypeBlocksArrayEqual(self,
            tb: TypeBlocks,
            match: tp.Iterable[object],
            match_dtype: tp.Optional[tp.Union[type, np.dtype, str]] = None) -> None:
        '''
        Args:
            tb: a TypeBlocks instance
            match: can be anything that can be used to create an array.
        '''
        # NOTE: this is comparing the potentially casted .values view, not each block
        # could use np.testing
        if not isinstance(match, np.ndarray):
            match = np.array(match, dtype=match_dtype)
        self.assertTrue((tb.values == match).all())


    def assertAlmostEqualValues(self,
            values1: tp.Iterable[object], values2: tp.Iterable[object]) -> None:

        for v1, v2 in zip_longest(values1, values2):
            self.assertEqualWithNaN(v1, v2)

    def assertAlmostEqualItems(self,
            pairs1: tp.Iterable[tp.Tuple[tp.Hashable, object]],
            pairs2: tp.Iterable[tp.Tuple[tp.Hashable, object]]) -> None:

        for (k1, v1), (k2, v2) in zip_longest(pairs1, pairs2):
            self.assertEqual(k1, k2)

            if isinstance(v1, float) and np.isnan(v1) and isinstance(v2, float) and np.isnan(v2):
                continue

            self.assertEqual(v1, v2)


    def assertAlmostEqualFramePairs(self,
            pairs1: tp.Iterable[tp.Tuple[tp.Hashable, tp.Iterable[object]]],
            pairs2: tp.Iterable[tp.Tuple[tp.Hashable, tp.Iterable[object]]]) -> None:
        '''
        For comparing nested tuples returned by Frame.to_pairs()
        '''
        # NOTE: this does not look at dtype or container classes
        for (k1, v1), (k2, v2) in zip_longest(pairs1, pairs2):
            self.assertEqual(k1, k2)
            self.assertAlmostEqualItems(v1, v2)


    def assertEqualFrames(self,
            f1: Frame,
            f2: Frame,
            compare_dtype: bool = True
            ) -> None:

        if not f1.equals(f2, compare_dtype=compare_dtype):
            self.assertTrue(f1.index.equals(f2.index, compare_dtype=compare_dtype), 'index do not match')
            self.assertTrue(f1.columns.equals(f2.columns, compare_dtype=compare_dtype), 'columns do not match')
            self.assertTrue(f1._blocks.equals(f2._blocks, compare_dtype=compare_dtype), '_blocks do not match')
            self.fail('class or name do not match')



    def assertEqualLines(self, lines1: str, lines2: str) -> None:
        '''After splitting and stripping, compare non-empty lines.
        '''
        def clean_lines(lines: str) -> tp.Iterator[str]:
            for line in lines.split('\n'):
                line = line.strip()
                if line:
                    yield line
        for l1, l2 in zip(clean_lines(lines1), clean_lines(lines2)):
            self.assertEqual(l1, l2)


# Helpful base types for testing
class UnHashable:
    '''UnHashable means __eq__ without defining __hash__'''

    def __init__(self, val: tp.Any) -> None:
        self.val = val

    def __eq__(self, other: tp.Any) -> bool:
        return hasattr(other, 'val') and self.val == other.val




import typing as tp
from datetime import datetime
from datetime import date

import numpy as np

from static_frame.core.node_selector import Interface
from static_frame.core.node_selector import TContainer
from static_frame.core.util import array_from_element_attr
from static_frame.core.util import array_from_element_method
from static_frame.core.util import array_from_element_apply
from static_frame.core.util import DT64_AS
from static_frame.core.util import DT64_DAY
from static_frame.core.util import DT64_FS
from static_frame.core.util import DT64_H
from static_frame.core.util import DT64_M
from static_frame.core.util import DT64_MONTH
from static_frame.core.util import DT64_MS
from static_frame.core.util import DT64_NS
from static_frame.core.util import DT64_PS
from static_frame.core.util import DT64_S
from static_frame.core.util import DT64_US
from static_frame.core.util import DT64_YEAR
from static_frame.core.util import DTYPE_DATETIME_KIND
from static_frame.core.util import DTYPE_INT_DEFAULT
from static_frame.core.util import DTYPE_OBJECT
from static_frame.core.util import DTYPE_STR
from static_frame.core.util import EMPTY_TUPLE
from static_frame.core.util import DTYPE_STR_KINDS

if tp.TYPE_CHECKING:

    from static_frame.core.frame import Frame  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index import Index  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.index_hierarchy import IndexHierarchy  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.series import Series  #pylint: disable = W0611 #pragma: no cover
    from static_frame.core.type_blocks import TypeBlocks  #pylint: disable = W0611 #pragma: no cover

BlocksType = tp.Iterable[np.ndarray]
ToContainerType = tp.Callable[[tp.Iterator[np.ndarray]], TContainer]

#https://docs.python.org/3/library/datetime.html

class InterfaceDatetime(Interface[TContainer]):

    __slots__ = (
            '_blocks', # function that returns iterable of arrays
            '_blocks_to_container', # partialed function that will return a new container
            )
    INTERFACE = (
            'year',
            'month',
            'day',
            'weekday',
            'timetuple',
            'fromisoformat',
            'isoformat',
            'strftime',
            'strptime',
            'strpdate',
            )

    DT64_EXCLUDE_YEAR = (DT64_YEAR,)
    DT64_EXCLUDE_YEAR_MONTH = (DT64_YEAR, DT64_MONTH)
    DT64_TIME = frozenset((
            DT64_H,
            DT64_M,
            DT64_S,
            DT64_MS,
            DT64_US,
            DT64_NS,
            DT64_PS,
            DT64_FS,
            DT64_AS,
            ))

    DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO = frozenset((
            DT64_YEAR,
            DT64_MONTH,
            DT64_NS,
            DT64_PS,
            DT64_FS,
            DT64_AS,
            ))

    def __init__(self,
            blocks: BlocksType,
            blocks_to_container: ToContainerType[TContainer]
            ) -> None:
        self._blocks: BlocksType = blocks
        self._blocks_to_container: ToContainerType[TContainer] = blocks_to_container

    @staticmethod
    def _validate_dtype_non_str(
            dtype: np.dtype,
            exclude: tp.Iterable[np.dtype] = EMPTY_TUPLE,
            ) -> None:
        '''
        Only support dtypes that are (or contain) datetime64 types. This is because most conversions from string can be done simply with astype().
        '''
        if ((dtype.kind == DTYPE_DATETIME_KIND
                or dtype == DTYPE_OBJECT)
                and dtype not in exclude
                ):
            return
        raise RuntimeError(f'invalid dtype ({dtype}) for date operation')


    @staticmethod
    def _validate_dtype_str(
            dtype: np.dtype,
            exclude: tp.Iterable[np.dtype] = EMPTY_TUPLE,
            ) -> None:
        '''
        Only support dtypes that are (or contain) strings.
        '''
        if ((dtype.kind in DTYPE_STR_KINDS
                or dtype == DTYPE_OBJECT)
                and dtype not in exclude
                ):
            return
        raise RuntimeError(f'invalid dtype ({dtype}) for operation on string types')

    #---------------------------------------------------------------------------

    @property
    def year(self) -> TContainer:
        'Return the year of each element.'

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    array = block.astype(DT64_YEAR).astype(DTYPE_INT_DEFAULT) + 1970
                    array.flags.writeable = False
                else: # must be object type
                    array = array_from_element_attr(
                            array=block,
                            attr_name='year',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    @property
    def month(self) -> TContainer:
        '''
        Return the month of each element, between 1 and 12 inclusive.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    array = block.astype(DT64_MONTH).astype(DTYPE_INT_DEFAULT) % 12 + 1
                    array.flags.writeable = False
                else: # must be object type
                    array = array_from_element_attr(
                            array=block,
                            attr_name='month',
                            dtype=DTYPE_INT_DEFAULT)
                yield array

        return self._blocks_to_container(blocks())

    @property
    def day(self) -> TContainer:
        '''
        Return the day of each element, between 1 and the number of days in the given month of the given year.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_DAY:
                        block = block.astype(DT64_DAY)
                    # subtract the first of the month, then shfit
                    array = (block - block.astype(DT64_MONTH)).astype(DTYPE_INT_DEFAULT) + 1
                    array.flags.writeable = False
                else: # must be object type
                    array = array_from_element_attr(
                            array=block,
                            attr_name='day',
                            dtype=DTYPE_INT_DEFAULT)

                yield array

        return self._blocks_to_container(blocks())



    #---------------------------------------------------------------------------

    # replace: akward to implement, as cannot provide None for the parameters that you do not want to set

    def weekday(self) -> TContainer:
        '''
        Return the day of the week as an integer, where Monday is 0 and Sunday is 6.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                self._validate_dtype_non_str(block.dtype, exclude=self.DT64_EXCLUDE_YEAR_MONTH)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype != DT64_DAY: # go to day first, then object
                        block = block.astype(DT64_DAY)
                    block = block.astype(DTYPE_OBJECT)
                # all object arrays by this point

                # returns an immutable array
                array = array_from_element_method(
                        array=block,
                        method_name='weekday',
                        args=EMPTY_TUPLE,
                        dtype=DTYPE_INT_DEFAULT
                        )
                yield array

        return self._blocks_to_container(blocks())


    #---------------------------------------------------------------------------
    # time methods

    def timetuple(self) -> TContainer:
        '''
        Return a ``time.struct_time`` such as returned by time.localtime().
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:

                # NOTE: nanosecond and lower will return integers; should exclude
                self._validate_dtype_non_str(block.dtype,
                        exclude=self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    block = block.astype(DTYPE_OBJECT)
                # all object arrays by this point

                # returns an immutable array
                array = array_from_element_method(
                        array=block,
                        method_name='timetuple',
                        args=EMPTY_TUPLE,
                        dtype=DTYPE_OBJECT
                        )
                yield array

        return self._blocks_to_container(blocks())

    def isoformat(self, sep: str = 'T', timespec: str = 'auto') -> TContainer:
        '''
        Return a string representing the date in ISO 8601 format, YYYY-MM-DD.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:

                self._validate_dtype_non_str(block.dtype,
                        exclude=self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO)

                args = EMPTY_TUPLE
                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    if block.dtype in self.DT64_TIME:
                        # if we know this is a time type, we can pass args
                        args = (sep, timespec) #type: ignore
                    block = block.astype(DTYPE_OBJECT)

                # all object arrays by this point
                # NOTE: we cannot determine if an Object array has date or datetime objects with a full iteration, so we cannot be sure if we need to pass args or not.

                # returns an immutable array
                array = array_from_element_method(
                        array=block,
                        method_name='isoformat',
                        args=args,
                        dtype=DTYPE_STR,
                        )
                yield array

        return self._blocks_to_container(blocks())

    def fromisoformat(self) -> TContainer:
        '''
        Return a :obj:`datetime.date` object from an ISO 8601 format.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit only string types, or objects types that contain strings
                self._validate_dtype_str(block.dtype)

                # NOTE: might use fromisoformat on date/datetime objects directly; assumed to be faster to go through datetime64 objects, fromisoformat is only available on python 3.7

                array_dt64 = block.astype(np.datetime64)
                if array_dt64.dtype in self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO:
                    raise RuntimeError(f'invalid derived dtype ({array_dt64.dtype}) for iso format')
                array = array_dt64.astype(object)
                array.flags.writeable = False
                yield array

        return self._blocks_to_container(blocks())

    def strftime(self, format: str) -> TContainer:
        '''
        Return a string representing the date, controlled by an explicit ``format`` string.
        '''

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:

                # NOTE: nanosecond and lower will return integers; should exclud
                self._validate_dtype_non_str(block.dtype,
                        exclude=self.DT64_EXCLUDE_YEAR_MONTH_SUB_MICRO)

                if block.dtype.kind == DTYPE_DATETIME_KIND:
                    block = block.astype(DTYPE_OBJECT)
                # all object arrays by this point

                # returns an immutable array
                array = array_from_element_method(
                        array=block,
                        method_name='strftime',
                        args=(format,),
                        dtype=DTYPE_STR,
                        )
                yield array

        return self._blocks_to_container(blocks())

    def strptime(self, format: str) -> TContainer:
        '''
        Return a Python datetime object from parsing a string defined with ``format``.
        '''
        def func(s: str) -> datetime:
            return datetime.strptime(s, format)

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit only string types, or objects types that contain strings
                self._validate_dtype_str(block.dtype)
                # returns an immutable array
                array = array_from_element_apply(
                        array=block,
                        func=func,
                        dtype=DTYPE_OBJECT,
                        )
                yield array

        return self._blocks_to_container(blocks())


    def strpdate(self, format: str) -> TContainer:
        '''
        Return a Python date object from parsing a string defined with ``format``.
        '''
        def func(s: str) -> date:
            return datetime.strptime(s, format).date()

        def blocks() -> tp.Iterator[np.ndarray]:
            for block in self._blocks:
                # permit only string types, or objects types that contain strings
                self._validate_dtype_str(block.dtype)
                # returns an immutable array
                array = array_from_element_apply(
                        array=block,
                        func=func,
                        dtype=DTYPE_OBJECT,
                        )
                yield array

        return self._blocks_to_container(blocks())
























import typing as tp

from static_frame.core.display_color import HexColor


if tp.TYPE_CHECKING:
    from static_frame.core.container import ContainerOperand #pylint: disable=W0611 #pragma: no cover

CSSDict = tp.Dict[str, str]


class StyleConfig:
    __slots__ = ('container',)

    def __init__(self, container: tp.Optional['ContainerOperand'] = None):
        self.container: tp.Optional['ContainerOperand'] = container

    def frame(self) -> str:
        '''
        Frame- (or table-) level styling.
        '''
        return ''

    def apex(self,
            value: tp.Any,
            coordinates: tp.Tuple[int, int],
            ) -> tp.Tuple[str, str]:
        '''
        Args:
            coordinates: negative integers to describe the apex space (i.e., the upper left corner formed when index and columns are displayed.
        Returns:
            A pair of the value to display, and a style information appropriate to the format.
        '''
        return str(value), ''

    def values(self,
            value: tp.Any,
            coordinates: tp.Tuple[int, int],
            ) -> tp.Tuple[str, str]:
        '''
        Returns:
            A pair of the value to display, and a style information appropriate to the format.
        '''
        return str(value), ''

    def index(self,
            label: tp.Hashable,
            ) -> tp.Tuple[str, str]:
        '''
        Returns:
            A pair of the value to display, and a style information appropriate to the format.
        '''
        return str(label), ''

    def columns(self,
            label: tp.Hashable,
            ) -> tp.Tuple[str, str]:
        '''
        Returns:
            A pair of the value to display, and a style information appropriate to the format.
        '''
        return str(label), ''


# Create an empty instance serve as default sentinal
STYLE_CONFIG_DEFAULT = StyleConfig()


class StyleConfigCSS(StyleConfig):

    COLOR_LIGHT_GREY = HexColor.get_html(0xe0e0e0)
    COLOR_OFF_BLACK = HexColor.get_html(0x2b2a2a)
    COLOR_GREY = HexColor.get_html(0xd1d2d4)
    COLOR_DARK_GREY = HexColor.get_html(0x898b8e)
    COLOR_WHITE = HexColor.get_html(0xffffff)
    COLOR_OFF_WHITE = HexColor.get_html(0xf2f2f2)

    FONT_SIZE = '14px'

    CSS_COMMON = dict(
            font_size=FONT_SIZE,
            border_width='1px',
            border_color=COLOR_DARK_GREY,
            border_style='solid',
            color=COLOR_OFF_BLACK,
    )

    @staticmethod
    def _dict_to_style(css_dict: CSSDict) -> str:
        '''
        Return a style attribute string containing all CSS in the CSSDict.
        '''
        style = ';'.join(f'{k.replace("_", "-")}:{v}' for k, v in css_dict.items())
        # NOTE: keep leading space to separate from tag
        return f' style="{style}"'


    def frame(self) -> str:
        '''
        Frame- (or table-) level styling.
        '''
        css = dict(
                border_collapse='collapse',
                border_width='1px',
                border_color=self.COLOR_DARK_GREY,
                border_style='solid',
                )
        return self._dict_to_style(css)

    def apex(self,
            value: tp.Any,
            coordinates: tp.Tuple[int, int],
            ) -> tp.Tuple[str, str]:

        css = dict(
                background_color=self.COLOR_LIGHT_GREY,
                font_weight='normal',
                padding='2px',
                border_width='0px',
                )
        return str(value), self._dict_to_style(css)

    def values(self,
            value: tp.Any,
            coordinates: tp.Tuple[int, int],
            ) -> tp.Tuple[str, str]:
        row, _ = coordinates

        def get_bg(row: int) -> str:
            if (row % 2) == 1:
                return self.COLOR_OFF_WHITE
            return self.COLOR_WHITE

        css = dict(
                background_color=get_bg(row),
                font_weight='normal',
                padding='2px',
                **self.CSS_COMMON,
                )
        return str(value), self._dict_to_style(css)

    def index(self,
            label: tp.Hashable,
            ) -> tp.Tuple[str, str]:
        css = dict(
                background_color=self.COLOR_GREY,
                font_weight='bold',
                **self.CSS_COMMON,
                )
        return str(label), self._dict_to_style(css)

    def columns(self,
            label: tp.Hashable,
            ) -> tp.Tuple[str, str]:

        css = dict(
                background_color=self.COLOR_GREY,
                font_weight='bold',
                **self.CSS_COMMON,
                )
        return str(label), self._dict_to_style(css)




def style_config_css_factory(
        style_config: tp.Optional[StyleConfig],
        container: 'ContainerOperand',
        ) -> tp.Optional[StyleConfig]:
    # let user set style_config to None to disable styling
    if style_config is STYLE_CONFIG_DEFAULT:
        # if given the base class, get the derived CSS class
        return StyleConfigCSS(container)
    if style_config is None:
        return None
    return style_config





'''Drop into an interactive interpreter pre-loaded with sf, np, and (if installed) pd.

$ python -m static_frame
$ ipython -m static_frame
'''


from code import interact
from sys import platform
from sys import version

import numpy as np

import static_frame as sf
from static_frame.core.display_color import HexColor

imports = {'np': np, 'sf': sf}

try:
    import pandas as pd
except ImportError: #pragma: no cover
    pass #pragma: no cover
else:
    imports['pd'] = pd

try:
    import frame_fixtures as ff
except ImportError: #pragma: no cover
    pass #pragma: no cover
else:
    imports['ff'] = ff

commands = sorted(
        f'import {package.__name__} as {name} # {package.__version__}'
        for name, package in imports.items()
        )

try: # This lets us play nicely with IPython:
    from builtins import __IPYTHON__  #type: ignore
    from IPython import embed
    from IPython import get_ipython

except ImportError:
    is_ipython = False
else:
    is_ipython = __IPYTHON__


if __name__ == '__main__':

    if is_ipython:
        ipython = get_ipython()

        print() # Spacer.
        for command in commands:
            ipython.auto_rewrite_input(command)
        print() # Spacer.

        embed(user_ns=imports, colors='neutral')

    else:
        banner_head = f'Python {version} on {platform}\n'
        banner_body = '\n'.join(f'>>> {command}' for command in commands)
        interact(banner=(banner_head + HexColor.format_terminal(0x505050, banner_body)),
                local=imports,
                exitmsg=''
                )
